@@ -56,6 +56,7 @@ static void message_add_recipients_from_string (GMimeMessage *message, char *typ
 static void message_set_subject (GMimeMessage *message, const char *subject);
 
 static ssize_t write_received (GMimeStream *stream, const char *name, const char *value);
+static ssize_t write_subject (GMimeStream *stream, const char *name, const char *value);
 static ssize_t write_msgid (GMimeStream *stream, const char *name, const char *value);
 
 
@@ -132,6 +133,7 @@ g_mime_message_init (GMimeMessage *message, GMimeMessageClass *klass)
 	message->message_id = NULL;
 	message->mime_part = NULL;
 	
+	g_mime_header_register_writer (((GMimeObject *) message)->headers, "Subject", write_subject);
 	g_mime_header_register_writer (((GMimeObject *) message)->headers, "Received", write_received);
 	g_mime_header_register_writer (((GMimeObject *) message)->headers, "Message-Id", write_msgid);
 }
@@ -182,7 +184,7 @@ typedef void (*token_skip_t) (const char **in);
 
 struct _received_token {
 	char *token;
-	int len;
+	size_t len;
 	token_skip_t skip;
 };
 
@@ -190,7 +192,7 @@ extern void decode_lwsp   (const char **in);
 
 static void skip_atom     (const char **in);
 static void skip_domain   (const char **in);
-static void skip_addrspec (const char **in);
+static void skip_addr     (const char **in);
 static void skip_msgid    (const char **in);
 
 static struct _received_token received_tokens[] = {
@@ -199,11 +201,9 @@ static struct _received_token received_tokens[] = {
 	{ "via ",  4, skip_atom     },
 	{ "with ", 5, skip_atom     },
 	{ "id ",   3, skip_msgid    },
-	{ "for ",  4, skip_addrspec }
+	{ "for ",  4, skip_addr     }
 };
 
-#define NUM_RECEIVED_TOKENS (sizeof (received_tokens) / sizeof (received_tokens[0]))
-
 static void
 skip_atom (const char **in)
 {
@@ -216,6 +216,29 @@ skip_atom (const char **in)
 	*in = inptr;
 }
 
+static void
+skip_comment (const char **in)
+{
+	register const char *inptr = *in;
+	int depth = 1;
+	
+	if (*inptr == '(')
+		inptr++;
+	
+	while (*inptr && depth > 0) {
+		if (*inptr == '(')
+			depth++;
+		else if (*inptr == ')')
+			depth--;
+		inptr++;
+	}
+	
+	if (*inptr == ')')
+		inptr++;
+	
+	*in = inptr;
+}
+
 static void
 skip_quoted_string (const char **in)
 {
@@ -336,6 +359,24 @@ skip_addrspec (const char **in)
 	*in = inptr;
 }
 
+static void
+skip_addr (const char **in)
+{
+	const char *inptr = *in;
+	
+	decode_lwsp (&inptr);
+	if (*inptr == '<') {
+		inptr++;
+		skip_addrspec (&inptr);
+		if (*inptr == '>')
+			inptr++;
+	} else {
+		skip_addrspec (&inptr);
+	}
+	
+	*in = inptr;
+}
+
 static void
 skip_msgid (const char **in)
 {
@@ -344,81 +385,126 @@ skip_msgid (const char **in)
 	decode_lwsp (&inptr);
 	if (*inptr == '<') {
 		inptr++;
-		decode_lwsp (&inptr);
 		skip_addrspec (&inptr);
 		if (*inptr == '>')
 			inptr++;
+	} else {
+		skip_atom (&inptr);
 	}
 	
 	*in = inptr;
 }
 
 
+struct _received_part {
+	struct _received_part *next;
+	const char *start;
+	size_t len;
+};
+
 static ssize_t
 write_received (GMimeStream *stream, const char *name, const char *value)
 {
-	const char *start, *inptr;
+	struct _received_part *parts, *part, *tail;
+	const char *lwsp, *inptr;
 	ssize_t nwritten;
 	GString *str;
 	int len, i;
 	
+	while (is_lwsp (*value))
+		value++;
+	
+	if (*value == '\0')
+		return 0;
+	
 	str = g_string_new (name);
 	g_string_append_len (str, ": ", 2);
 	len = 10;
 	
-	start = inptr = value;
+	tail = parts = part = g_alloca (sizeof (struct _received_part));
+	part->start = inptr = value;
+	part->next = NULL;
+	
 	while (*inptr) {
-		while (is_lwsp (*inptr))
-			inptr++;
-		
-		for (i = 0; i < NUM_RECEIVED_TOKENS; i++) {
+		for (i = 0; i < G_N_ELEMENTS (received_tokens); i++) {
 			if (!strncmp (inptr, received_tokens[i].token, received_tokens[i].len)) {
-				if ((inptr - start) + len > GMIME_FOLD_LEN && start != value) {
-					g_string_append (str, "\n\t");
-					while (is_lwsp (*start))
-						start++;
-					len = 1;
+				if (inptr > part->start) {
+					part->len = lwsp - part->start;
+					
+					part = g_alloca (sizeof (struct _received_part));
+					part->start = inptr;
+					part->next = NULL;
+					
+					tail->next = part;
+					tail = part;
 				}
 				
-				/* write the last section */
-				g_string_append_len (str, start, inptr - start);
-				len += (inptr - start);
-				start = inptr;
-				
 				inptr += received_tokens[i].len;
 				received_tokens[i].skip (&inptr);
-				decode_lwsp (&inptr);
-				inptr--;
+				
+				lwsp = inptr;
+				while (is_lwsp (*inptr))
+					inptr++;
+				
+				if (*inptr == ';') {
+					inptr++;
+					
+					part->len = inptr - part->start;
+					
+					lwsp = inptr;
+					while (is_lwsp (*inptr))
+						inptr++;
+					
+					part = g_alloca (sizeof (struct _received_part));
+					part->start = inptr;
+					part->next = NULL;
+					
+					tail->next = part;
+					tail = part;
+				}
 				
 				break;
 			}
 		}
 		
-		if (*inptr == ';') {
-			if ((inptr - start) + len > GMIME_FOLD_LEN && start != value) {
-				g_string_append (str, "\n\t");
-				while (is_lwsp (*start))
-					start++;
-				len = 1;
-			}
+		if (i == G_N_ELEMENTS (received_tokens)) {
+			while (*inptr && !is_lwsp (*inptr))
+				inptr++;
 			
-			inptr++;
-			g_string_append_len (str, start, inptr - start);
-			len += (inptr - start);
-			start = inptr;
+			lwsp = inptr;
+			while (is_lwsp (*inptr))
+				inptr++;
 		}
 		
-		inptr++;
+		if (*inptr == '(') {
+			skip_comment (&inptr);
+			
+			lwsp = inptr;
+			while (is_lwsp (*inptr))
+				inptr++;
+		}
 	}
 	
-	if ((inptr - start) + len > GMIME_FOLD_LEN && start != value) {
-		g_string_append (str, "\n\t");
-		while (is_lwsp (*start))
-			start++;
-		len = 1;
-	}
+	part->len = lwsp - part->start;
+	
+	lwsp = NULL;
+	part = parts;
+	do {
+		len += lwsp ? part->start - lwsp : 0;
+		if (len + part->len > GMIME_FOLD_LEN && part != parts) {
+			g_string_append (str, "\n\t");
+			len = 1;
+		} else if (lwsp) {
+			g_string_append_len (str, lwsp, part->start - lwsp);
+		}
+		
+		g_string_append_len (str, part->start, part->len);
+		lwsp = part->start + part->len;
+		len += part->len;
+		
+		part = part->next;
+	} while (part != NULL);
 	
-	g_string_append_len (str, start, inptr - start);
 	g_string_append_c (str, '\n');
 	
 	nwritten = g_mime_stream_write (stream, str->str, str->len);
@@ -427,6 +513,22 @@ write_received (GMimeStream *stream, const char *name, const char *value)
 	return nwritten;
 }
 
+static ssize_t
+write_subject (GMimeStream *stream, const char *name, const char *value)
+{
+	char *unfolded, *folded;
+	ssize_t n;
+	
+	unfolded = g_strdup_printf ("%s: %s\n", name, value);
+	folded = g_mime_utils_unstructured_header_fold (unfolded);
+	g_free (unfolded);
+	
+	n = g_mime_stream_write_string (stream, folded);
+	g_free (folded);
+	
+	return n;
+}
+
 static ssize_t
 write_msgid (GMimeStream *stream, const char *name, const char *value)
 {
@@ -657,23 +759,20 @@ message_write_to_stream (GMimeObject *object, GMimeStream *stream)
 	ssize_t nwritten, total = 0;
 	
 	/* write the content headers */
-	nwritten = g_mime_header_write_to_stream (object->headers, stream);
-	if (nwritten == -1)
+	if ((nwritten = g_mime_header_write_to_stream (object->headers, stream)) == -1)
 		return -1;
 	
 	total += nwritten;
 	
 	if (message->mime_part) {
-		nwritten = g_mime_stream_write_string (stream, "MIME-Version: 1.0\n");
-		if (nwritten == -1)
+		if ((nwritten = g_mime_stream_write_string (stream, "MIME-Version: 1.0\n")) == -1)
 			return -1;
 		
 		total += nwritten;
 		
 		nwritten = g_mime_object_write_to_stream (message->mime_part, stream);
 	} else {
-		nwritten = g_mime_stream_write (stream, "\n", 1);
-		if (nwritten == -1)
+		if ((nwritten = g_mime_stream_write (stream, "\n", 1)) == -1)
 			return -1;
 	}
 	
@@ -728,16 +827,25 @@ message_set_sender (GMimeMessage *message, const char *sender)
  * @sender: The name and address of the sender
  *
  * Set the sender's name and address on the MIME Message.
- * (ex: "\"Joe Sixpack\" <joe@sixpack.org>")
+ * (ex: "\"Joe Sixpack\" &lt;joe@sixpack.org&gt;")
  **/
 void
 g_mime_message_set_sender (GMimeMessage *message, const char *sender)
 {
+	InternetAddressList *addrlist;
+	char *encoded;
+	
 	g_return_if_fail (GMIME_IS_MESSAGE (message));
 	g_return_if_fail (sender != NULL);
 	
 	message_set_sender (message, sender);
-	g_mime_header_set (GMIME_OBJECT (message)->headers, "From", message->from);
+	
+	addrlist = internet_address_parse_string (sender);
+	encoded = internet_address_list_to_string (addrlist, TRUE);
+	internet_address_list_destroy (addrlist);
+	
+	g_mime_header_set (GMIME_OBJECT (message)->headers, "From", encoded);
+	g_free (encoded);
 }
 
 
@@ -1142,6 +1250,28 @@ g_mime_message_get_header (GMimeMessage *message, const char *header)
 }
 
 
+/**
+ * g_mime_message_get_mime_part:
+ * @message: MIME Message
+ *
+ * Gets the toplevel MIME part contained within @message.
+ *
+ * Returns the toplevel MIME part of @message.
+ **/
+GMimeObject *
+g_mime_message_get_mime_part (GMimeMessage *message)
+{
+	g_return_val_if_fail (GMIME_IS_MESSAGE (message), NULL);
+	
+	if (message->mime_part == NULL)
+		return NULL;
+	
+	g_object_ref (message->mime_part);
+	
+	return message->mime_part;
+}
+
+
 /**
  * g_mime_message_set_mime_part:
  * @message: MIME Message
@@ -1170,6 +1300,9 @@ g_mime_message_set_mime_part (GMimeMessage *message, GMimeObject *mime_part)
  *
  * Write the contents of the MIME Message to @stream.
  *
+ * WARNING: This interface is deprecated. Use
+ * g_mime_object_write_to_stream() instead.
+ *
  * Returns -1 on fail.
  **/
 ssize_t
@@ -1178,7 +1311,7 @@ g_mime_message_write_to_stream (GMimeMessage *message, GMimeStream *stream)
 	g_return_val_if_fail (GMIME_IS_MESSAGE (message), -1);
 	g_return_val_if_fail (GMIME_IS_STREAM (stream), -1);
 	
-	return g_mime_object_write_to_stream (GMIME_OBJECT (message), stream);
+	return g_mime_object_write_to_stream ((GMimeObject *) message, stream);
 }
 
 
@@ -1188,6 +1321,9 @@ g_mime_message_write_to_stream (GMimeMessage *message, GMimeStream *stream)
  *
  * Allocates a string buffer containing the mime message @message.
  *
+ * WARNING: This interface is deprecated. Use
+ * g_mime_object_to_string() instead.
+ *
  * Returns an allocated string containing the MIME Message.
  **/
 char *
@@ -1195,12 +1331,11 @@ g_mime_message_to_string (GMimeMessage *message)
 {
 	g_return_val_if_fail (GMIME_IS_MESSAGE (message), NULL);
 	
-	return g_mime_object_to_string (GMIME_OBJECT (message));
+	return g_mime_object_to_string ((GMimeObject *) message);
 }
 
 
-/**
- * The proper way to handle a multipart/alternative part is to return
+/* The proper way to handle a multipart/alternative part is to return
  * the last part that we know how to render. For our purposes, we are
  * going to assume:
  *
@@ -1263,7 +1398,7 @@ handle_multipart_mixed (GMimeMultipart *multipart, gboolean want_plain, gboolean
 		} else if (g_mime_content_type_is_type (type, "text", "*")) {
 			if (!strcasecmp (type->subtype, want_plain ? "plain" : "html")) {
 				/* we got what we came for */
-				*is_html = !want_plain;
+				*is_html = !strcasecmp (type->subtype, "html");
 				return mime_part;
 			}
 			
@@ -1325,10 +1460,10 @@ g_mime_message_get_body (const GMimeMessage *message, gboolean want_plain, gbool
 			mime_part = handle_multipart_mixed (multipart, want_plain, is_html);
 	} else if (g_mime_content_type_is_type (type, "text", "*")) {
 		/* this *has* to be the message body */
-		if (g_mime_content_type_is_type (type, "text", want_plain ? "plain" : "html"))
-			*is_html = !want_plain;
+		if (g_mime_content_type_is_type (type, "text", "html"))
+			*is_html = TRUE;
 		else
-			*is_html = want_plain;
+			*is_html = FALSE;
 		
 		mime_part = message->mime_part;
 	}
