@@ -25,6 +25,7 @@
 #include <config.h>
 #endif
 
+#include <stdlib.h>
 #include <string.h>
 #include <sys/types.h>
 #include <regex.h>
@@ -357,7 +358,7 @@ g_mime_parser_new (void)
 {
 	GMimeParser *parser;
 	
-	parser = g_object_new (GMIME_TYPE_PARSER, NULL, NULL);
+	parser = g_object_new (GMIME_TYPE_PARSER, NULL);
 	
 	return parser;
 }
@@ -398,11 +399,11 @@ g_mime_parser_new_with_stream (GMimeStream *stream)
  * or resetting of the stream. Anything that will/could change the
  * current stream's offset is PROHIBITED.
  *
- * It is also recommended that you not use #g_mime_stream_tell because
- * it will not necessarily give you the current @parser offset since
- * @parser handles its own internal read-ahead buffer. Instead, it is
- * recommended that you use #g_mime_parser_tell if you have a reason
- * to need the current offset of the @parser.
+ * It is also recommended that you not use g_mime_stream_tell()
+ * because it will not necessarily give you the current @parser offset
+ * since @parser handles its own internal read-ahead buffer. Instead,
+ * it is recommended that you use g_mime_parser_tell() if you have a
+ * reason to need the current offset of the @parser.
  **/
 void
 g_mime_parser_init_with_stream (GMimeParser *parser, GMimeStream *stream)
@@ -624,6 +625,9 @@ parser_fill (GMimeParser *parser)
 static off_t
 parser_offset (struct _GMimeParserPrivate *priv, unsigned char *inptr)
 {
+	if (priv->offset == -1)
+		return -1;
+	
 	if (!inptr)
 		inptr = priv->inptr;
 	
@@ -828,10 +832,13 @@ parser_step_headers (GMimeParser *parser)
 			}
 			
 			/* check to see if we've reached the end of the headers */
-			if (!priv->midline && inptr == start)
+			if (!priv->midline && (inptr == start || (inptr - start == 1 && *start == '\r')))
 				goto headers_end;
 			
 			len = inptr - start;
+			if (inptr[-1] == '\r')
+				len--;
+			
 			header_backup (priv, start, len);
 			
 			if (inptr < inend) {
@@ -969,6 +976,9 @@ check_boundary (struct _GMimeParserPrivate *priv, const unsigned char *start, si
 {
 	off_t offset = parser_offset (priv, (unsigned char *) start);
 	
+	if (start[len - 1] == '\r')
+		len--;
+	
 	if (possible_boundary (priv->scan_from, start, len)) {
 		struct _boundary_stack *s;
 		
@@ -1013,7 +1023,7 @@ check_boundary (struct _GMimeParserPrivate *priv, const unsigned char *start, si
  **/
 
 static int
-parser_scan_content (GMimeParser *parser, GByteArray *content)
+parser_scan_content (GMimeParser *parser, GByteArray *content, int *crlf)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
 	register unsigned char *inptr;
@@ -1057,9 +1067,10 @@ parser_scan_content (GMimeParser *parser, GByteArray *content)
 			len = inptr - start;
 			
 			if (inptr < inend) {
-				inptr++;
 				if ((found = check_boundary (priv, start, len)))
 					goto boundary;
+				
+				inptr++;
 				len++;
 			} else {
 				/* didn't find an end-of-line */
@@ -1088,6 +1099,15 @@ parser_scan_content (GMimeParser *parser, GByteArray *content)
 	/* don't chew up the boundary */
 	priv->inptr = start;
 	
+	if (found != FOUND_EOS) {
+		if (inptr[-1] == '\r')
+			*crlf = 2;
+		else
+			*crlf = 1;
+	} else {
+		*crlf = 0;
+	}
+	
 	return found;
 }
 
@@ -1100,19 +1120,20 @@ parser_scan_mime_part_content (GMimeParser *parser, GMimePart *mime_part, int *f
 	GMimeDataWrapper *wrapper;
 	GMimeStream *stream;
 	off_t start, end;
+	int crlf;
 	
 	if (priv->persist_stream && priv->seekable)
 		start = parser_offset (priv, NULL);
 	else
 		content = g_byte_array_new ();
 	
-	*found = parser_scan_content (parser, content);
+	*found = parser_scan_content (parser, content, &crlf);
 	if (*found != FOUND_EOS) {
 		/* last '\n' belongs to the boundary */
 		if (priv->persist_stream && priv->seekable)
-			end = parser_offset (priv, NULL) - 1;
+			end = parser_offset (priv, NULL) - crlf;
 		else
-			g_byte_array_set_size (content, MAX (content->len - 1, 0));
+			g_byte_array_set_size (content, MAX (content->len - crlf, 0));
 	} else if (priv->persist_stream && priv->seekable) {
 		end = parser_offset (priv, NULL);
 	}
@@ -1152,12 +1173,11 @@ parser_scan_message_part (GMimeParser *parser, GMimeMessagePart *mpart, int *fou
 		header = header->next;
 	}
 	
-	content_type = parser_content_type (parser);
-	if (!content_type)
+	if (!(content_type = parser_content_type (parser)))
 		content_type = g_mime_content_type_new ("text", "plain");
 	
 	parser_unstep (parser);
-	if (content_type && g_mime_content_type_is_type (content_type, "multipart", "*")) {
+	if (g_mime_content_type_is_type (content_type, "multipart", "*")) {
 		object = parser_construct_multipart (parser, content_type, found);
 	} else {
 		object = parser_construct_leaf_part (parser, content_type, found);
@@ -1181,12 +1201,6 @@ parser_construct_leaf_part (GMimeParser *parser, GMimeContentType *content_type,
 	while (parser_step (parser) != GMIME_PARSER_STATE_HEADERS_END)
 		;
 	
-	if (!content_type) {
-		content_type = parser_content_type (parser);
-		if (!content_type)
-			content_type = g_mime_content_type_new ("text", "plain");
-	}
-	
 	object = g_mime_object_new_type (content_type->type, content_type->subtype);
 	header = priv->headers;
 	while (header) {
@@ -1196,7 +1210,9 @@ parser_construct_leaf_part (GMimeParser *parser, GMimeContentType *content_type,
 	
 	header_raw_clear (&priv->headers);
 	
-	g_mime_object_set_content_type (object, content_type);
+	if (object->content_type)
+		g_mime_content_type_destroy (object->content_type);
+	object->content_type = content_type;
 	
 	/* skip empty line after headers */
 	parser_skip_line (parser);
@@ -1210,27 +1226,53 @@ parser_construct_leaf_part (GMimeParser *parser, GMimeContentType *content_type,
 	return object;
 }
 
+static void
+crlf2lf (unsigned char *in)
+{
+	register unsigned char *inptr = in;
+	register unsigned char *outptr;
+	
+	while (*inptr != '\0' && !(inptr[0] == '\r' && inptr[1] == '\n'))
+		inptr++;
+	
+	if (*inptr == '\0')
+		return;
+	
+	outptr = inptr++;
+	
+	while (*inptr != '\0') {
+		while (*inptr != '\0' && !(inptr[0] == '\r' && inptr[1] == '\n'))
+			*outptr++ = *inptr++;
+		
+		if (*inptr == '\r')
+			inptr++;
+	}
+	
+	*outptr = '\0';
+}
+
 static int
 parser_scan_multipart_face (GMimeParser *parser, GMimeMultipart *multipart, gboolean preface)
 {
 	GByteArray *buffer;
 	const char *face;
-	int found;
+	int crlf, found;
 	
 	buffer = g_byte_array_new ();
-	found = parser_scan_content (parser, buffer);
+	found = parser_scan_content (parser, buffer, &crlf);
 	
-	/* last '\n' belongs to the boundary */
-	if (buffer->len) {
-		buffer->data[buffer->len - 1] = '\0';
+	if (buffer->len > crlf) {
+		/* last '\n' belongs to the boundary */
+		g_byte_array_set_size (buffer, buffer->len + 1);
+		buffer->data[buffer->len - crlf - 1] = '\0';
+		crlf2lf (buffer->data);
 		face = buffer->data;
-	} else
-		face = NULL;
-	
-	if (preface)
-		g_mime_multipart_set_preface (multipart, face);
-	else
-		g_mime_multipart_set_postface (multipart, face);
+		
+		if (preface)
+			g_mime_multipart_set_preface (multipart, face);
+		else
+			g_mime_multipart_set_postface (multipart, face);
+	}
 	
 	g_byte_array_free (buffer, TRUE);
 	
@@ -1254,8 +1296,7 @@ parser_scan_multipart_subparts (GMimeParser *parser, GMimeMultipart *multipart)
 		/* get the headers */
 		parser_step_headers (parser);
 		
-		content_type = parser_content_type (parser);
-		if (!content_type)
+		if (!(content_type = parser_content_type (parser)))
 			content_type = g_mime_content_type_new ("text", "plain");
 		
 		parser_unstep (parser);
@@ -1302,7 +1343,9 @@ parser_construct_multipart (GMimeParser *parser, GMimeContentType *content_type,
 	
 	header_raw_clear (&priv->headers);
 	
-	g_mime_object_set_content_type (object, content_type);
+	if (object->content_type)
+		g_mime_content_type_destroy (object->content_type);
+	object->content_type = content_type;
 	
 	multipart = (GMimeMultipart *) object;
 	
@@ -1346,8 +1389,7 @@ parser_construct_part (GMimeParser *parser)
 	while (parser_step (parser) != GMIME_PARSER_STATE_HEADERS_END)
 		;
 	
-	content_type = parser_content_type (parser);
-	if (!content_type)
+	if (!(content_type = parser_content_type (parser)))
 		content_type = g_mime_content_type_new ("text", "plain");
 	
 	parser_unstep (parser);
@@ -1409,8 +1451,7 @@ parser_construct_message (GMimeParser *parser)
 			priv->bounds->content_end = parser_offset (priv, NULL) + content_length;
 	}
 	
-	content_type = parser_content_type (parser);
-	if (!content_type)
+	if (!(content_type = parser_content_type (parser)))
 		content_type = g_mime_content_type_new ("text", "plain");
 	
 	parser_unstep (parser);
@@ -1454,7 +1495,7 @@ g_mime_parser_construct_message (GMimeParser *parser)
  * @parser: MIME parser object
  *
  * Gets the mbox-style From-line of the most recently parsed message
- * (gotten from #g_mime_parser_construct_message).
+ * (gotten from g_mime_parser_construct_message()).
  *
  * Returns the mbox-style From-line of the most recently parsed
  * message or %NULL on error.
@@ -1482,7 +1523,7 @@ g_mime_parser_get_from (GMimeParser *parser)
  * @parser: MIME parser object
  *
  * Gets the offset of the most recently parsed mbox-style From-line
- * (gotten from #g_mime_parser_construct_message).
+ * (gotten from g_mime_parser_construct_message()).
  *
  * Returns the offset of the most recently parsed mbox-style From-line
  * or -1 on error.
