@@ -981,16 +981,8 @@ g_mime_references_clear (GMimeReferences **refs)
 }
 
 
-/**
- * g_mime_utils_header_fold:
- * @in: input header string
- *
- * Folds a header according to the rules in rfc822.
- *
- * Returns an allocated string containing the folded header.
- **/
-char *
-g_mime_utils_header_fold (const char *in)
+static char *
+header_fold (const char *in, gboolean structured)
 {
 	gboolean last_was_lwsp = FALSE;
 	register const char *inptr;
@@ -1009,11 +1001,16 @@ g_mime_utils_header_fold (const char *in)
 		len = strcspn (inptr, " \t\n");
 		
 		if (outlen + len > GMIME_FOLD_LEN) {
-			if (last_was_lwsp)
-				g_string_insert_c (out, out->len - 1, '\n');
-			else
-				g_string_append (out, "\n\t");
-			outlen = 1;
+			if (outlen > 1) {
+				if (last_was_lwsp) {
+					if (structured)
+						out->str[out->len - 1] = '\t';
+					
+					g_string_insert_c (out, out->len - 1, '\n');
+				} else
+					g_string_append (out, "\n\t");
+				outlen = 1;
+			}
 			
 			/* check for very long words, just cut them up */
 			while (outlen + len > GMIME_FOLD_LEN) {
@@ -1031,6 +1028,7 @@ g_mime_utils_header_fold (const char *in)
 			inptr += len;
 			last_was_lwsp = FALSE;
 		} else {
+			last_was_lwsp = TRUE;
 			if (*inptr == '\t') {
 				/* tabs are a good place to fold, odds
 				   are that this is where the previous
@@ -1038,11 +1036,9 @@ g_mime_utils_header_fold (const char *in)
 				g_string_append (out, "\n\t");
 				outlen = 1;
 				inptr++;
-				last_was_lwsp = FALSE;
 			} else {
 				g_string_append_c (out, *inptr++);
 				outlen++;
-				last_was_lwsp = TRUE;
 			}
 		}
 	}
@@ -1057,6 +1053,51 @@ g_mime_utils_header_fold (const char *in)
 }
 
 
+/**
+ * g_mime_utils_structured_header_fold:
+ * @in: input header string
+ *
+ * Folds a structured header according to the rules in rfc822.
+ *
+ * Returns an allocated string containing the folded header.
+ **/
+char *
+g_mime_utils_structured_header_fold (const char *in)
+{
+	return header_fold (in, TRUE);
+}
+
+
+/**
+ * g_mime_utils_unstructured_header_fold:
+ * @in: input header string
+ *
+ * Folds an unstructured header according to the rules in rfc822.
+ *
+ * Returns an allocated string containing the folded header.
+ **/
+char *
+g_mime_utils_unstructured_header_fold (const char *in)
+{
+	return header_fold (in, FALSE);
+}
+
+
+/**
+ * g_mime_utils_header_fold:
+ * @in: input header string
+ *
+ * Folds a structured header according to the rules in rfc822.
+ *
+ * Returns an allocated string containing the folded header.
+ **/
+char *
+g_mime_utils_header_fold (const char *in)
+{
+	return header_fold (in, TRUE);
+}
+
+
 /**
  * g_mime_utils_header_printf:
  * @format: string format
@@ -1078,7 +1119,7 @@ g_mime_utils_header_printf (const char *format, ...)
 	buf = g_strdup_vprintf (format, ap);
 	va_end (ap);
 	
-	ret = g_mime_utils_header_fold (buf);
+	ret = header_fold (buf, TRUE);
 	g_free (buf);
 	
 	return ret;
@@ -1759,7 +1800,7 @@ rfc2047_encode_get_rfc822_words (const unsigned char *in, gboolean phrase)
 		
 		inptr = newinptr;
 		
-		if (g_unichar_isspace (c)) {
+		if (c < 256 && is_lwsp (c)) {
 			if (count > 0) {
 				word = g_new (struct _rfc822_word, 1);
 				word->next = NULL;
@@ -1789,6 +1830,23 @@ rfc2047_encode_get_rfc822_words (const unsigned char *in, gboolean phrase)
 				type = WORD_2047;
 				encoding = 2;
 			}
+			
+			if (count >= GMIME_FOLD_PREENCODED) {
+				word = g_new (struct _rfc822_word, 1);
+				word->next = NULL;
+				word->start = start;
+				word->end = inptr;
+				word->type = type;
+				word->encoding = encoding;
+				
+				tail->next = word;
+				tail = word;
+				count = 0;
+				
+				start = inptr;
+				type = WORD_ATOM;
+				encoding = 0;
+			}
 		}
 		
 		last = inptr;
@@ -1994,8 +2052,8 @@ g_mime_utils_header_encode_text (const unsigned char *in)
  *
  * Decodes an rfc2047 encoded header.
  *
- * Note: This function is deprecated. Use
- * #g_mime_utils_header_decode_text instead.
+ * WARNING: This function is deprecated. Use
+ * g_mime_utils_header_decode_text() instead.
  *
  * Returns the decoded header (which will be in UTF-8 if at all
  * possible).
@@ -2013,8 +2071,8 @@ g_mime_utils_8bit_header_decode (const unsigned char *in)
  *
  * Encodes a 'text' header according to the rules in rfc2047.
  *
- * Note: This function is deprecated. Use
- * #g_mime_utils_header_encode_text instead.
+ * WARNING: This function is deprecated. Use
+ * g_mime_utils_header_encode_text() instead.
  *
  * Returns the encoded header. Useful for encoding
  * headers like "Subject".
@@ -2032,8 +2090,8 @@ g_mime_utils_8bit_header_encode (const unsigned char *in)
  *
  * Encodes a 'phrase' header according to the rules in rfc2047.
  *
- * Note: This function is deprecated. Use
- * #g_mime_utils_header_encode_phrase instead.
+ * WARNING: This function is deprecated. Use
+ * g_mime_utils_header_encode_phrase() instead.
  *
  * Returns the encoded 'phrase'. Useful for encoding internet
  * addresses.
@@ -2054,7 +2112,7 @@ g_mime_utils_8bit_header_encode_phrase (const unsigned char *in)
  * @save: leftover bits that have not yet been encoded
  *
  * Base64 encodes the input stream to the output stream. Call this
- * when finished encoding data with #g_mime_utils_base64_encode_step
+ * when finished encoding data with g_mime_utils_base64_encode_step()
  * to flush off the last little bit.
  *
  * Returns the number of bytes encoded.
@@ -2255,7 +2313,7 @@ g_mime_utils_base64_decode_step (const unsigned char *in, size_t inlen, unsigned
  * @save: leftover bits that have not yet been encoded
  *
  * Uuencodes a chunk of data. Call this when finished encoding data
- * with #g_mime_utils_uuencode_step to flush off the last little bit.
+ * with g_mime_utils_uuencode_step() to flush off the last little bit.
  *
  * Returns the number of bytes encoded.
  **/
@@ -2516,7 +2574,7 @@ g_mime_utils_uudecode_step (const unsigned char *in, size_t inlen, unsigned char
  * @save: leftover bits that have not yet been encoded
  *
  * Quoted-printable encodes a block of text. Call this when finished
- * encoding data with #g_mime_utils_quoted_encode_step to flush off
+ * encoding data with g_mime_utils_quoted_encode_step() to flush off
  * the last little bit.
  *
  * Returns the number of bytes encoded.
